from .fire_and_forget import fire_and_forget
from threading import Thread
from typing import Any, Callable, List


class Promise:
    def __init__(self, coroutine: Callable, *params: List[Any]):
        def coroutine_wrapper():
            self.returned_value = coroutine(*params)
            return self.returned_value

        self.returned_value: Any = None
        self.coroutine: Callable = coroutine_wrapper
        self.params: List[Any] = params

    @staticmethod
    def all(*promises: List["Promise"]):
        threads = [promise.to_thread() for promise in promises]
        # start all threads
        [thread.start() for thread in threads]
        # wait for all threads to complete
        [thread.join() for thread in threads]

        return [promise.returned_value for promise in promises]

    def then(self, callback: Callable, error: Callable = lambda: None) -> None:
        @fire_and_forget
        def cb():
            try:
                value = self.coroutine()
                callback(value)
            except Exception as ex:
                if (error != None):
                    error(ex)
                else:
                    raise ex
        cb()

    def to_thread(self) -> Thread:
        return Thread(target=self.coroutine)

    def wait(self):
        self.to_thread().run()
        return self.returned_value



from pypromises.promise import Promise
import time


def task_uploading(sec):
    time.sleep(sec)
    return "Uploading Ended"


def task_test(sec):
    time.sleep(sec)
    return "Testing ended"


def error_task(message, secs):
    time.sleep(secs)
    raise Exception(message)


promise_upload = Promise(task_uploading, 1)
promise_testing = Promise(task_test, 0.5)
promise_error = Promise(error_task, "Hahah", 0.5)

promise_upload.then(lambda value: print(value))
promise_testing.then(lambda value: print(value))
promise_error.then(lambda value: print(value), lambda err: print(err))



from pypromises.promise import Promise
import time


def task_uploading(sec):
    time.sleep(sec)
    return "Uploading Ended"


def task_test(sec):
    time.sleep(sec)
    return "Testing ended"


promise_upload = Promise(task_uploading, 0.5)
promise_testing = Promise(task_test, 1)

start = time.time()
print(Promise.all(promise_upload, promise_testing))
print(time.time()-start)



import asyncio


def fire_and_forget(f):
    def wrapped(*args, **kwargs):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return asyncio.get_event_loop().run_in_executor(None, f, *args, *kwargs)

    return wrapped
